/* This code tests a general MPI file view
 * where each process write/read a number
 * of elements equal to its rank, assuming
 * no information is provided on how many
 * elements should be read/written.
 *
 * write patter should be:
 * 0 1 1 2 2 2 ... N (N-times) 0 1 1 2 2 2 ... 
 */
#include <stdlib.h>
#include <stdio.h>

#include "mpi.h"

#define PARALLEL_IO_PATH_TO_FILE "./output_c.dat"
#define DIM_BUFF 10

int main( int argc, char *argv[] )
{
    MPI_Init( &argc, &argv );

    int world_size, myrank;
    MPI_Comm_size(MPI_COMM_WORLD, &world_size);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    MPI_File     fh;
    MPI_Datatype filetype;
    MPI_Status   f_status;

    int *count_proc = (int *) malloc (world_size * sizeof(int));
    for (int i = 0; i < world_size; i++) {
      count_proc[i] = i + 1;
    }

    MPI_Aint ub, lb;
    int etype_extent;

    MPI_Datatype cont_t;
    MPI_Type_contiguous(count_proc[myrank], MPI_INT, &cont_t);
    MPI_Type_commit(&cont_t);


    int *count_disp = (int *) malloc (world_size * sizeof(int));
    etype_extent = 0;
    count_disp[0] = 0;
    for (int i = 0; i < world_size; i++) {
      count_disp[i] = etype_extent;
      etype_extent += i + 1;
    }

    int size_int;
    MPI_Type_size(MPI_INT, &size_int);

    MPI_Type_create_resized(cont_t, 0, etype_extent * size_int, &filetype);
    MPI_Type_commit(&filetype);

    MPI_File_delete(PARALLEL_IO_PATH_TO_FILE, MPI_INFO_NULL);
    MPI_File_open(MPI_COMM_WORLD, PARALLEL_IO_PATH_TO_FILE,
      MPI_MODE_CREATE | MPI_MODE_WRONLY,  MPI_INFO_NULL, &fh);
    MPI_Offset offset = (MPI_Offset) count_disp[myrank] * size_int;

    MPI_File_set_view(fh, offset, MPI_INT, filetype, "native", MPI_INFO_NULL);

    int my_dim_buf = count_proc[myrank] * DIM_BUFF;
    int *buf =  (int *) malloc (my_dim_buf * sizeof(int));
    for (int i = 0; i < my_dim_buf; i++) {
      buf[i] = myrank;
    }

    MPI_File_write(fh, buf, my_dim_buf, MPI_INT, &f_status);

    MPI_File_close(&fh);

    MPI_Finalize();

    return 0;
}
