#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#define min(A,B) (A<B?A:B)
int fill_matrix(int gsizes[2], int i, int j) {
  return i+j;
}

#include <sys/time.h>
double get_time()
{
    struct timeval current;
    gettimeofday(&current , NULL);
    return (double)current.tv_sec*1000000 + (double)current.tv_usec;
}

#include "mpi.h"

#define PARALLEL_IO_PATH_TO_FILE "./output_c.dat"
#define ind(i,j) ((size_t)i*lsizes[1]+j)

int main( int argc, char *argv[] )
{
//  Declare variables (or do it later)
    int m = 17; // rows of local matrix
    int n = 13; // cols of local matrix

//  Start MPI
    MPI_Init( &argc, &argv );

    if (argc != 3) {
      int rank;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      if (rank == 0) {
        printf("\n%s nx ny\n", argv[0]);
        printf("where:\n nx ny are local grid size per process along x and y\n\n");
      }
      exit(EXIT_FAILURE);
    }
    else 
    {
      m = atoi(argv[1]); 
      n = atoi(argv[2]); 
    }

//  Set cartesian topology
    int world_size;
    MPI_Comm_size(MPI_COMM_WORLD, &world_size);

    int dims[2]; dims[0] = 0; dims[1] = 0;
    MPI_Dims_create(world_size, 2, dims);

    int periods[2]; periods[0] = 0; periods[1] = 0;

    MPI_Comm     comm;
    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, 1, &comm);

    int rank, coords[2];
    MPI_Comm_rank(comm, &rank);
    MPI_Cart_coords(comm, rank, 2, coords);


    // set subarray info
    int rem;
    int gsizes[2], psizes[2];
    int lsizes[2], start_indices[2];

    psizes[0] = dims[0];  /* no. of processes in vertical dimension  of process grid */
    psizes[1] = dims[1];  /* no. of processes in horizontal dimension  of process grid */

    gsizes[0] = m * dims[0];  /* no. of rows in global array */
    gsizes[1] = n * dims[1];  /* no. of columns in global array*/

    lsizes[0] = m;   /* no. of rows in local array */
    lsizes[1] = n;   /* no. of rows in local array */

    start_indices[0] = coords[0] * lsizes[0];
    start_indices[1] = coords[1] * lsizes[1];

    // lsizes[0] = gsizes[0]/psizes[0];   /* no. of rows in local array */
    // rem = gsizes[0]%psizes[0];
    // if (rem && coords[0] < rem) {
    //   lsizes[0]++;
    //   start_indices[0] = coords[0] * lsizes[0];
    // } else {
    //   start_indices[0] = rem + coords[0] * lsizes[0];
    // }

    // lsizes[1] = gsizes[1]/psizes[1];   /* no. of columns in local array */
    // rem = gsizes[1]%psizes[1];
    // if (rem && coords[1] < rem) {
    //   lsizes[1]++;
    //   start_indices[1] = coords[1] * lsizes[1];
    // } else {
    //   start_indices[1] = rem + coords[1] * lsizes[1];
    // }

    if (rank == 0) {
      printf("\nUsing a grid of [%d][%d] processes\n",
          dims[0], dims[1]);
      double memusage = (double) lsizes[0] * lsizes[1] * sizeof(int) / (1024*1024);
      double totfilesize = (double) dims[0] * dims[1] * memusage;
      printf("Memory per process: %lf [MB], Total file size %lf [MB]\n",
        memusage, totfilesize);
    }

    // initialize local matrix (array) 
    size_t local_array_size = (size_t) lsizes[0] * lsizes[1];
    if (local_array_size < 0) {
      fprintf(stderr,"\nERROR: local_array_size is negative\n");
      fprintf(stderr,"\tlsizes[0]=%d lsizes[1]=%d local_array_size=%zd\n",
        lsizes[0], lsizes[1], local_array_size);
    }

    int *local_array = (int *) malloc( local_array_size * sizeof(int) );
    for (int i=0; i<lsizes[0]; i++) 
        for (int j=0; j<lsizes[1]; j++) {
            // this won't work if i*j> max int representation
            // int value = gsizes[1]*(i+start_indices[0]+1)+(j+start_indices[1]+1);
            int value = fill_matrix(gsizes, i+start_indices[0]+1, j+start_indices[1]+1);
            local_array[ind(i,j)] = value;
        }

    /* some timing variables */
    double start_us, stop_us;

    /* create subarray filetype for MPI File view */
    MPI_File     fh;
    MPI_Datatype filetype;
    MPI_Status   f_status;

//  Create subarray, open and write file in parallel
    MPI_Type_create_subarray(2, gsizes, lsizes, start_indices,
                         MPI_ORDER_C, MPI_INT, &filetype);
    MPI_Type_commit(&filetype);

    MPI_File_delete(PARALLEL_IO_PATH_TO_FILE, MPI_INFO_NULL);
    MPI_File_open(MPI_COMM_WORLD, PARALLEL_IO_PATH_TO_FILE,
      MPI_MODE_CREATE | MPI_MODE_WRONLY,  MPI_INFO_NULL, &fh);
    MPI_Offset displ = 0;
    MPI_File_set_view(fh, displ, MPI_INT, filetype, "native",  MPI_INFO_NULL);

start_us = get_time();
{
  size_t total = 0;
  int mpi_file_io_chunk_size = INT_MAX / sizeof(int);
  do {
    int mpi_file_ask_chunk = min(local_array_size - total,  mpi_file_io_chunk_size);
    int count;
    // MPI_File_write_all(fh, local_array + total, mpi_file_ask_chunk,
    MPI_File_write(fh, local_array + total, mpi_file_ask_chunk,
           MPI_INT, &f_status);
    MPI_Get_count(&f_status, MPI_INT, &count);
    total += count;
    fprintf(stderr,"rank %d written %zd [bytes]\n", rank, total*sizeof(int));
  } while(total < local_array_size);
}
stop_us = get_time();
    printf("MPI_File_write: rank %4d  time %20lf\n", rank, (stop_us-start_us)*1.E-6 );

    MPI_File_close(&fh);
stop_us = get_time();

    printf("MPI_File_write+MPI_File_close: rank %4d  time %20lf\n", rank, (stop_us-start_us)*1.E-6 );

#ifdef VERIFY_PARALLEL
    // Parallel verify 
    int errcount = 0;
    int *verify_array = (int *) calloc( local_array_size , sizeof(int) );

start_us = get_time();
    MPI_File_open(MPI_COMM_WORLD, PARALLEL_IO_PATH_TO_FILE,
      MPI_MODE_RDONLY,  MPI_INFO_NULL, &fh);
    MPI_File_set_view(fh, displ, MPI_INT, filetype, "native",  MPI_INFO_NULL);
{
  size_t total = 0;
  int mpi_file_io_chunk_size = INT_MAX / sizeof(int);
  do {
    int mpi_file_ask_chunk = min(local_array_size - total,  mpi_file_io_chunk_size);
    int count;
    // MPI_File_read_all(fh, verify_array + total, mpi_file_ask_chunk,
    MPI_File_read(fh, verify_array + total, mpi_file_ask_chunk,
           MPI_INT, &f_status);
    MPI_Get_count(&f_status, MPI_INT, &count);
    total += count;
    fprintf(stderr,"rank %d read    %zd [bytes]\n", rank, total*sizeof(int));
  } while(total < local_array_size);
}
stop_us = get_time();

    printf("MPI_File_read:  rank %4d  time %20lf\n", rank, (stop_us-start_us)*1.E-6 );


    for (int i=0; i<lsizes[0]; i++) 
        for (int j=0; j<lsizes[1]; j++) 
            if (verify_array[(i,j)] != local_array[(i,j)]) {
               fprintf(stderr,"Parallel Verify ERROR: at indexes %d,%d read=%d, written=%d \n",
                       i, j, verify_array[ind(i,j)], local_array[ind(i,j)]);
               errcount++;
            }
    if ( errcount == 0 ) {
      printf("Parallel Verify: test passed on proc(%d)\n", rank);
    }

    MPI_File_close(&fh);

    MPI_Type_free(&filetype);
    free(local_array);
    free(verify_array);
#endif // VERIFY_PARALLEL

#ifdef VERIFY_SERIAL
    // Serial verify 
    errcount = 0;
    if(rank == 0) {
       FILE* fh;
       fh = fopen(PARALLEL_IO_PATH_TO_FILE,"rb");
       int *serial_verify_array = (int *) calloc(gsizes[1] , sizeof(int) );
       for (int i=0; i<gsizes[0]; i++) {
           fread(serial_verify_array,sizeof(int),gsizes[1],fh);
           for (int j=0; j<gsizes[1]; j++) {
               int value = fill_matrix(gsizes, i+1, j+1);
               if (serial_verify_array[j] != value) {
                   fprintf(stderr,"Serial Verify ERROR: at index %d,%d read=%d, written=%d \n",
                          i, j, serial_verify_array[j], value);
                  errcount++;
               }
           }
       }

       if (!errcount) {
         printf("Serial Verify: test passed\n");
       }

       free(serial_verify_array);
       fclose(fh);
    }
#endif // VERIFY_SERIAL


//  MPI Finalize
    MPI_Finalize();

    return 0;
}
